/*
 * record_mpeg2ps: "records" MPEG2 Program Streams
 *   (i.e. copies data from the input to one or more output files,
 *   with the ability to split the stream on a group-of-pictures
 *   border and to filter out invalid data)
 * 
 * Copyright (C) 2004 Christian Wolff
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * Or, point your browser to http://www.gnu.org/copyleft/gpl.html
 * 
 * The author can be reached at sub_recps@scarabaeus.org, 
 * the project's page is at http://scara.com/recps/
 * 
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dirent.h>
#include <malloc.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <curses.h>
#include <unistd.h>
#include <limits.h>
#include <time.h>

//#undef NULL
//#define NULL (void *)0

int PTS_to_time(unsigned long PTS, int *hour, int *minute, int *second, int *frame)
{
	if (!(hour && minute && second)) return 1;
	if (frame) {
		PTS /= 900;
		*frame = PTS % 100;
		PTS /= 100;
	} else {
		PTS /= 90000;
	}
	*second = PTS % 60;
	PTS /= 60;
	*minute = PTS % 60;
	PTS /= 60;
	*hour = PTS;
	return 0;
}

unsigned short CRC;

int had_pack_header = 0;
unsigned long long last_pack_header = 0;
unsigned long last_pack_SCR_base = 0;
unsigned int last_pack_SCR_ext = 0;
unsigned long last_PTS = 0;
int had_system_header = 0;
int had_program_end_code = 0;

int need_break = 0;
// User interface:
int force_new = 0;  // start new file if 1
int stop_record = 0;  // stop recording and close file if 1
int pause_record = 0;  // go into pause
int unpause_record = 0;  // leave pause mode
int pause_mode = 0;
int intermission_mode = 0;

int pipethrough = 0;
int overwrite_seg = 0;
unsigned long long maxsize = 0;
int use_timestamp = 0;

char filebase[1024];
char timestamp[23];
int segment = 0;

int groupstart = 0;

int video_analyze (unsigned char * data, int length, unsigned long long position)
{
	int p;
	
	p = 3;
	while (p < length) {
	
		// Search for start code
		if ((data[p - 1] != 0x01) || data[p - 2] || data[p - 3]) p++;
		else {
			if (data[p++] == 0xB3) {
				groupstart = 1;
				return 0;
			}
		}
	}
	return 0;
}


// analyzes PES packet ('length' bytes in 'data'), 'position' holds the 
// absolut position in the file/stream of the first byte of the start code
int pes_analyze(unsigned char * data, int length, unsigned long long position)
{
	int i, n;
	
	unsigned int stream_id, PES_packet_length, PES_header_data_length;
	int PTS_flag = 0;
	unsigned long PTS = 0;
	
	int marker_bits = 1;
	
	stream_id = data[3] & 0xFF;
	PES_packet_length = ((data[4] & 0x00FF) << 8) | (data[5] & 0x00FF);
	
	if (
		(stream_id >= 0xBC) && 
		(stream_id != 0xBC) &&
		(stream_id != 0xBE) &&
		(stream_id != 0xBF) &&
		(stream_id != 0xF0) &&
		(stream_id != 0xF1) &&
		(stream_id != 0xF2) &&
		(stream_id != 0xF8) &&
		(stream_id != 0xFF)
	) {  // PES-Header, parse for subheaders
		marker_bits &= ((data[6] & 0xC0) == 0x80);
		PTS_flag = data[7] & 0x80;
		PES_header_data_length = data[8] & 0xFF;
		
		n = 9;
		if (PTS_flag) {
			PTS = (data[n++] >> 1) & 0x07ULL;                      // Bit 32-30
			PTS = (PTS << 8) | (data[n++] & 0xFFULL);         // Bit 29-22
			PTS = (PTS << 7) | ((data[n++] >> 1) & 0x7FULL);  // Bit 21-15
			PTS = (PTS << 8) | (data[n++] & 0xFFULL);         // Bit 14-7
			PTS = (PTS << 7) | ((data[n++] >> 1) & 0x7FULL);  // Bit 6-0
		}
	}
	
	if (marker_bits) {
		if (PTS_flag) 
			last_PTS = PTS;
		if ((data[3] & 0xF0) == 0xE0) {  // video
			i = PES_header_data_length + 9;
			video_analyze(&data[i], PES_packet_length - (PES_header_data_length + 3), position + i);
		}
	}
	
	return marker_bits;
}

int pack_header_analyze(unsigned char * data, int length, unsigned long long position)
{
	unsigned long SCR_base;
//	unsigned int SCR_ext;
//	unsigned long mux_rate;
	int marker_bits;
	
	// interpret header data
	marker_bits = (
		((data[4] & 0xC4) == 0x44) && 
		(data[6] & 0x04) && 
		(data[8] & 0x04) && 
		(data[9] & 0x01) && 
		((data[12] & 0x03) == 0x03) && 
		((data[13] & 0xF8) == 0xF8) 
	);
	
	SCR_base = (data[4] >> 3) & 0x07UL;  // Bit 32-30
	SCR_base = (SCR_base << 2) | (data[4] & 0x03UL);  // Bit 29-28
	SCR_base = (SCR_base << 8) | (data[5] & 0xFFUL);  // Bit 27-20
	SCR_base = (SCR_base << 5) | ((data[6] >> 3) & 0x1FUL);  // Bit 19-15
	SCR_base = (SCR_base << 2) | (data[6] & 0x03UL);  // Bit 14-13
	SCR_base = (SCR_base << 8) | (data[7] & 0xFFUL);  // Bit 12-5
	SCR_base = (SCR_base << 5) | ((data[8] >> 3) & 0x7FUL);  // Bit 4-0
//	SCR_ext = data[8] & 0x03; // Bit 8-7
//	SCR_ext = (SCR_ext << 7) | ((data[9] >> 1) & 0x7F);  // Bit 6-0
//	mux_rate = data[10];
//	mux_rate = (mux_rate << 8) | data[11];
//	mux_rate = (mux_rate << 6) | ((data[12] >> 2) & 0x3FUL);
	
	if (marker_bits) {
		last_pack_SCR_base = SCR_base;
//		last_pack_SCR_ext = SCR_ext;
	}
	
	return marker_bits;
}

int system_header_analyze(unsigned char * data, int length, unsigned long long position)
{
	int marker_bits;
	int j;
	
	// interpret header data
	marker_bits = (
		(data[6] & 0x80) && 
		(data[8] & 0x01) && 
		(data[10] & 0x20) && 
		((data[11] & 0x7F) == 0x7F) 
	);
	
	j = 12;
	while (((j + 2) < length) && (data[j] & 0x80)) {
		marker_bits &= ((data[j + 1] & 0xC0) == 0xC0);
		j += 3;
	}
	
	return marker_bits;
}

void update_led(void) 
{
	if (stop_record) {
		// TODO CapsLock LED off, ScrollLock off
	} else if (intermission_mode) {
		// TODO CapsLock LED off, ScrollLock on
	} else {
		// TODO CapsLock LED on, ScrollLock off
	}
}

int read_stream(char * filename)
{
	FILE* inputstream;
	FILE* outputstream;
	int i, junk, packetsize, missing, alldata;
	char outfilename[1024];
	unsigned long long processed = 0;
	unsigned long long written = 0;
	unsigned long long junkstart = 0;
	unsigned long long datastart = 0;
	static unsigned char startcode[] = {0x00, 0x00, 0x01};
	static unsigned char program_end_code[] = {0x00, 0x00, 0x01, 0xB9};
	unsigned char *data;
	unsigned char maindata[65542];  // 64k + 6 byte
	int marker_bits;
	int was_pack_header = 0;
	int h, m, s;
	int key;
	int no_split = 0;  // 1 if output is stdout, files can't be split
	struct stat filestat;
	
	data = maindata;
	
	inputstream = stdin;
	if (strcmp(filename, "-")) {
		if ((inputstream = fopen(filename, "r")) == NULL) {
			fprintf(stderr,"\rERROR: Couldn't open file %s for reading\r\n", filename);
			return 1;
		}
	}
	
	if (pipethrough) {
		if (strcmp(filebase, "-")) {
			if (use_timestamp) {
				time_t now = time(NULL);
				strftime(timestamp, 23, "_%Y-%m-%d_%H-%M-%S", localtime(&now));
			}
			if (maxsize) {
				sprintf(outfilename, "%s%s_%02d.mpg", filebase, timestamp, segment);
				if (!overwrite_seg) {
					while (stat(outfilename, &filestat) >= 0) {
//						if (S_ISREG(filestat.st_mode)) {  // directory? create separate FIFO's
						segment++;
						sprintf(outfilename, "%s%s_%02d.mpg", filebase, timestamp, segment);
					}
				}
			} else {
				sprintf(outfilename, "%s%s.mpg", filebase, timestamp);
			}
			if ((outputstream = fopen(outfilename, "w")) == NULL) {
				fprintf(stderr,"\rERROR: Couldn't open file %s for writing\r\n", outfilename);
				return 1;
			} else {
				fprintf(stderr,"\rWriting to file %s\r\n", outfilename);
			}
		} else {
			outputstream = stdout;
			maxsize = 0;
			no_split = 1;
		}
	}
	
	junk = 0;
	missing = 0;
	alldata = 0;
	while (!feof(inputstream) && !ferror(inputstream)) {
		if ((key = getch()) != ERR) {
			switch (key) {
				case 'r':  // start new file ([r]ecord)
				case 'R':
					if (intermission_mode) {
						if (!no_split) force_new = 1;
						intermission_mode = 0;
						update_led();
					}
					break;
				case 's':  // stop file, but keep recording in intermediate file ([s]top)
				case 'S':
					if (! intermission_mode) {
						if (!no_split) force_new = 1;
						intermission_mode = 1;
						update_led();
					}
					break;
				case 'n':  // start new file, toggle intermediate status ([n]ew)
				case 'N':
					if (!no_split) force_new = 1;
					intermission_mode = ! intermission_mode;
					update_led();
					break;
					
				case 'q':  // stop and exit ([q]uit)
				case 'Q':
					stop_record = 1;
					update_led();
					break;
				case 'p':  // pause toggle ([p]ause)
				case 'P':
					if (pause_mode) unpause_record = 1;
					else pause_record = 1;
					break;
				case 'c':  // continue after pause ([c]ontinue)
				case 'C':
					unpause_record = 1;
					break;
				case 'h':  // [h]elp
				case 'H':
				case '?':
					fprintf(stderr,"Keys: [q]uit, [r]ecord, [s]top, [n]ew, [p]ause, [c]ontinue, [h]elp\n");
					fprintf(stderr,"  [q]uit:     stop recording at the next group-of-pictures border and exit program\n");
					fprintf(stderr,"  [r]ecord:   stop recording the current file, start a new file in 'recording' mode\n");
					fprintf(stderr,"  [s]top:     stop recording the current file, start a new file in 'intermission' mode\n");
					fprintf(stderr,"  [n]ew:      stop recording the current file, start a new file and toggle 'rec.'/'int.' mode\n");
					fprintf(stderr,"  [p]ause:    discard stream from next GOP on, or resume when in pause already\n");
					fprintf(stderr,"  [c]ontinue: resume recoding after pause\n");
					break;
			}
		}
		
		// chop off recent junk
		if (junk || missing) {
			junk = 0;
			missing = 0;
			junkstart = processed;
		}
		
		// Search for start code
		i = 0;
		while ((i < 3) && ! feof(inputstream) && ! ferror(inputstream)) {
			if (fread(&data[i], 1, 1, inputstream) <= 0) {
				missing = 1;
				continue;
			}
			junk++;
			processed++;
			if (data[i] == startcode[i]) i++;
			else if (data[i]) i = 0;  // reset counter only if third byte is not zero.
		}
		if (i < 3) {
			continue;
		}
		
		if (junk > 3) {
			junk = 3;
			junkstart = processed - 3;
		}
		
		// Read stream ID
		if (fread(&data[i++], 1, 1, inputstream) <= 0) {
			missing = 1;
			continue;
		}
		junk++;
		processed++;
		datastart = processed - 4;
		
		if (data[3] == 0xBA) {  // pack header, 10 byte plus up to 7 byte padding
			if (fread(&data[i], 10, 1, inputstream) <= 0) {
				missing = 10;
				continue;
			}
			i+=10;
			junk+=10;
			processed+=10;
			
			if ((marker_bits = pack_header_analyze(data, i, datastart))) {
				had_pack_header = 1;
				was_pack_header = 1;
				last_pack_header = datastart;
				
				if (data[13] & 0x07) {  // read padding bytes, if any
					if (fread(&data[i], data[13] & 0x07, 1, inputstream) <= 0 ) {
						missing = data[13] & 0x07;
						continue;
					}
					i += (data[13] & 0x07);
					junk += (data[13] & 0x07);
					processed += (data[13] & 0x07);
				}
				
				// pass valid data through to stdout
				if (pipethrough) {
					//fwrite(data, i, 1, stdout);
					//written += i;
					alldata += i;
					data += i;
				}
				
				junk -= i;
			}
		} else if (was_pack_header && (data[3] == 0xBB)) {  // system header
			if (fread(&data[i], 2, 1, inputstream) <= 0) {
				missing = 2;
				continue; 
			}
			i += 2;
			junk += 2;
			processed += 2;
			packetsize = ((data[4] & 0x00FF) << 8) | (data[5] & 0x00FF);
			if (fread(&data[i], packetsize, 1, inputstream) <= 0) {
				missing = packetsize;
				continue;
			}
			i += packetsize;
			junk += packetsize;
			processed += packetsize;
			
			marker_bits = system_header_analyze(data, i, datastart);
			
			had_system_header = 1;
			was_pack_header = 0;
			
			// pass valid data through to stdout
			if (marker_bits) {
				if (pipethrough) {
					//fwrite(&data[0], i, 1, stdout);
					//written += i;
					alldata += i;
					data += i;
				}
				junk -= i;
			}
		} else if (had_pack_header && (data[3] == 0xB9)) {  // program end code
			had_program_end_code = 1;
			
			// pass valid data through to stdout
			if (pipethrough) {
				//fwrite(&data[0], i, 1, stdout);
				//written += i;
				alldata += i;
				data += i;
				if (!pause_mode) fwrite(&maindata[0], alldata, 1, outputstream);
				written += alldata;
				alldata = 0;
				data = maindata;
			}
			if (stop_record) {
				PTS_to_time(last_PTS, &h, &m, &s, NULL);
				fprintf(stderr, "\r*** Closing file at program end code: position: %10lld, %02d:%02d:%02d\r\n",
					last_pack_header,
					h, m, s);
				fclose(outputstream);
				stop_record = 0;
				return 0;
			}
			junk -= i;
			
			was_pack_header = 0;
		} else if (had_pack_header && (data[3] >= 0xBC)) {  // PES packet
			if (fread(&data[i], 2, 1, inputstream) <= 0) {
				missing = 2;
				continue; 
			}
			i += 2;
			junk += 2;
			processed += 2;
			packetsize = ((data[4] & 0x00FF) << 8) | (data[5] & 0x00FF);
			if (fread(&data[i], packetsize, 1, inputstream) <= 0) {
				missing = packetsize;
				continue;
			}
			i += packetsize;
			junk += packetsize;
			processed += packetsize;
			
			groupstart = 0;
			marker_bits = pes_analyze(data, i, datastart);
			
			// pass valid data through to stdout
			if (marker_bits) {
				if (groupstart) {
					PTS_to_time(last_PTS, &h, &m, &s, NULL);
					fprintf(stderr, "\rTimeCode: %02d:%02d:%02d", h, m, s);
					if (stop_record) {
						PTS_to_time(last_PTS, &h, &m, &s, NULL);
						fprintf(stderr, "\r*** Closing file: position: %10lld, %02d:%02d:%02d\r\n",
							last_pack_header,
							h, m, s);
						fwrite(program_end_code, sizeof(program_end_code), 1, outputstream);
						fclose(outputstream);
						stop_record = 0;
						return 0;
					}
					if (force_new) {
						PTS_to_time(last_PTS, &h, &m, &s, NULL);
						if (use_timestamp) {
							time_t now = time(NULL);
							strftime(timestamp, 23, "_%Y-%m-%d_%H-%M-%S", localtime(&now));
						}
						fprintf(stderr, "\r*** New file: position: %10lld, %02d:%02d:%02d\r\n",
							last_pack_header,
							h, m, s);
						segment = 0;
						if (maxsize) {
							sprintf(outfilename, "%s%s_%02d.mpg", filebase, timestamp, segment);
						} else {
							sprintf(outfilename, "%s%s.mpg", filebase, timestamp);
						}
						fclose(outputstream);
						if ((outputstream = fopen(outfilename, "w")) == NULL) {
							fprintf(stderr,"\rERROR: Couldn't open file %s for writing\r\n", outfilename);
							return 1;
						} else {
							fprintf(stderr,"\rWriting to file %s\r\n", outfilename);
						}
						written = 0;
						force_new = 0;
					}
					if (maxsize && (written >= maxsize)) {
						PTS_to_time(last_PTS, &h, &m, &s, NULL);
						fprintf(stderr, "\r*** New segment: position: %10lld, %02d:%02d:%02d\r\n",
							last_pack_header,
							h, m, s);
						segment++;
						sprintf(outfilename, "%s%s_%02d.mpg", filebase, timestamp, segment);
						fclose(outputstream);
						if ((outputstream = fopen(outfilename, "w")) == NULL) {
							fprintf(stderr,"\rERROR: Couldn't open file %s for writing\r\n", outfilename);
							return 1;
						} else {
							fprintf(stderr,"\rWriting to file %s\r\n", outfilename);
						}
						written = 0;
					}
					if (pause_record) {
						PTS_to_time(last_PTS, &h, &m, &s, NULL);
						fprintf(stderr, "\r*** Pause: position: %10lld, %02d:%02d:%02d\r\n",
							last_pack_header,
							h, m, s);
						pause_mode = 1;
						pause_record = 0;
					}
					if (unpause_record) {
						PTS_to_time(last_PTS, &h, &m, &s, NULL);
						fprintf(stderr, "\r*** Continue: position: %10lld, %02d:%02d:%02d\r\n",
							last_pack_header,
							h, m, s);
						pause_mode = 0;
						unpause_record = 0;
					}
				}
				if (pipethrough) {
					//fwrite(&data[0], i, 1, stdout);
					//written += i;
					alldata += i;
					data += i;
					if (!pause_mode) fwrite(&maindata[0], alldata, 1, outputstream);
					written += alldata;
					alldata = 0;
					data = maindata;
				}
				junk -= i;
			}
			was_pack_header = 0;
		} else {  //invalid header
			was_pack_header = 0;
		}
	}
	
	return 0;
}

int main(int argc, char* argv[])
{
	int i;
	int filearg = 0;
	
	if (argc < 2) {
		fprintf(stderr, "record_mpeg2ps by Christian Wolff.\n\
Usage: %s [-m <bytes>] [-p <output file>] [-o] [-t] <input file>\n\
-m: maxmum file size (plus some kbyte)\n\
-p: pipe filtered data through to stdout ('-') or <file>_00.mpg etc.\n\
-o: overwrite existing segment files\n\
-t: include starting time stamp in filename\n\
<file>: filename or '-' for stdin\n\
",argv[0]);
		exit (0);
	}
	
	i = 1;
	while (i < argc) {
		if (argv[i][0] == '-') {  // option
			if (argv[i][1] == 'm') {
				if (strlen(argv[i]) > 2) {
					maxsize = atoi(&argv[i][2]);
				} else if ((i + 1) < argc) {
					i++;
					maxsize = atoi(argv[i]);
				}
			} else if (argv[i][1] == 'p') {
				pipethrough = 1;  // pipethrough on
				if (strlen(argv[i]) > 2) {
					strncpy(filebase, &argv[i][2], 1024);
				} else if ((i + 1) < argc) {
					i++;
					strncpy(filebase, argv[i], 1024);
				}
			} else if (argv[i][1] == 'o') {
				overwrite_seg = 1;
			} else if (argv[i][1] == 't') {
				use_timestamp = 1;
			}
		} else if (!filearg) {
			filearg = i;
		}
		i++;
	}
	timestamp[0] = '\0';
	
	// curses intialisation		
	initscr(); 
	refresh();
	cbreak(); 
	noecho();
	nonl();
	intrflush(stdscr, FALSE);
	keypad(stdscr, TRUE);
	nodelay(stdscr, TRUE);
	
	update_led();
	
	if (maxsize) fprintf(stderr, "Maximum file size: %lld byte\r\n", maxsize);
	if (pipethrough) fprintf(stderr, "Piping to file: %s\r\n", filebase);
	
	if (filearg) read_stream(argv[filearg]);
	else read_stream("-");
	
	// curses de-init
	clear();
	refresh();
	endwin();
	
	return 0;
}
